@@ -11,6 +11,7 @@ use Algorithm::Diff qw(sdiff);
 our @EXPORT_OK = qw(
     diff
     list_diff
+    split_diff
     dsplit
     dtraverse
     patch
@@ -30,15 +31,15 @@ Struct::Diff - Recursive diff tools for nested perl structures
 
 =head1 VERSION
 
-Version 0.87
+Version 0.88
 
 =cut
 
-our $VERSION = '0.87';
+our $VERSION = '0.88';
 
 =head1 SYNOPSIS
 
-    use Struct::Diff qw(diff dsplit list_diff patch);
+    use Struct::Diff qw(diff list_diff split_diff patch);
 
     $a = {x => [7,{y => 4}]};
     $b = {x => [7,{y => 9}],z => 33};
@@ -46,13 +47,13 @@ our $VERSION = '0.87';
     $diff = diff($a, $b, noO => 1, noU => 1); # omit unchanged items and old values
     # $diff == {D => {x => {D => [{I => 1,N => {y => 9}}]},z => {A => 33}}}
 
-    $splitted = dsplit($diff);
-    # $splitted->{a} # not exists
-    # $splitted->{b} == {x => [{y => 9}],z => 33}
-
     @list_diff = list_diff($diff); # list (path and ref pairs) all diff entries
     # $list_diff == [[{keys => ['z']}],\{A => 33},[{keys => ['x']},[0]],\{I => 1,N => {y => 9}}]
 
+    $splitted = split_diff($diff);
+    # $splitted->{a} # not exists
+    # $splitted->{b} == {x => [{y => 9}],z => 33}
+
     patch($a, $diff); # $a now equal to $b by structure and data
 
 =head1 EXPORT
@@ -131,13 +132,12 @@ sub diff($$;@) {
     if (ref $a ne ref $b) {
         $d->{O} = $a unless ($opts{noO});
         $d->{N} = $b unless ($opts{noN});
-    } elsif (ref $a eq 'ARRAY') {
-        return $opts{noU} ? {} : { U => $a } if ($a == $b);
-
+    } elsif (ref $a eq 'ARRAY' and $a != $b) {
         my @sd = sdiff($a, $b, sub { freeze \$_[0] });
-        my ($s, $hidden);
+        my ($s, $hidden, $item);
+
         for my $i (0 .. $#sd) {
-            my $item;
+            undef $item;
             if ($sd[$i]->[0] eq 'u') {
                 $item->{U} = $sd[$i]->[1] unless ($opts{noU});
             } elsif ($sd[$i]->[0] eq 'c') {
@@ -158,16 +158,14 @@ sub diff($$;@) {
             }
         }
 
-        if ((my @k = keys %{$s}) == 1 and not ($hidden or exists $s->{D})) { # all have same status - return it
+        if ((my @k = keys %{$s}) == 1 and not ($hidden or exists $s->{D})) { # all items have same status
             map { $_ = $_->{$k[0]} } @{$d->{D}};
             $d->{$k[0]} = delete $d->{D};
         }
 
         $d->{U} = $a unless ($hidden or keys %{$d});
-    } elsif (ref $a eq 'HASH') {
-        return $opts{noU} ? {} : { U => $a } if ($a == $b);
-
-        my @keys = keys %{{ %{$a}, %{$b} }}; # united uniq keys
+    } elsif (ref $a eq 'HASH' and $a != $b) {
+        my @keys = keys %{{ %{$a}, %{$b} }}; # uniq keys for both hashes
         return $opts{noU} ? {} : { U => {} } unless (@keys);
 
         my ($alt, $sd);
@@ -181,7 +179,9 @@ sub diff($$;@) {
                     if (exists $sd->{D}) {
                         $d->{D}->{$key} = $alt->{D}->{$key} = $sd;
                     } else {
-                        map { $d->{$_}->{$key} = $alt->{D}->{$key}->{$_} = $sd->{$_} } keys %{$sd};
+                        map {
+                            $d->{$_}->{$key} = $alt->{D}->{$key}->{$_} = $sd->{$_}
+                        } keys %{$sd};
                     }
                 }
             } elsif (exists $a->{$key}) {
@@ -205,6 +205,15 @@ sub diff($$;@) {
     return $d;
 }
 
+=head2 dsplit
+
+Is an alias for L</split_diff>. Deprecated, will be removed in future
+releases. L</split_diff> should be used instead.
+
+=cut
+
+*dsplit = \&split_diff;
+
 =head2 list_diff
 
 List pairs (path, ref_to_subdiff) for provided diff. See
@@ -243,33 +252,42 @@ sub list_diff($;@) {
         if (!exists ${$diff}->{D} or $opts{depth} and @{$path} >= $opts{depth}) {
             push @list, $path, $diff;
         } elsif (ref ${$diff}->{D} eq 'ARRAY') {
-            map { unshift @stack, [@{$path}, [$_]], \${$diff}->{D}->[$_] }
-                reverse 0 .. $#{${$diff}->{D}};
+            map {
+                unshift @stack,
+                    [@{$path},
+                        [exists ${$diff}->{D}->[$_]->{I}
+                            ? ${$diff}->{D}->[$_]->{I} # use provided index
+                            : $_
+                        ]
+                    ],
+                    \${$diff}->{D}->[$_]
+            } reverse 0 .. $#{${$diff}->{D}};
         } else { # HASH
-            map { unshift @stack, [@{$path}, {keys => [$_]}], \${$diff}->{D}->{$_} }
-                $opts{sort}
-                    ? (ref $opts{sort} eq 'CODE'
-                        ? reverse $opts{sort}(keys %{${$diff}->{D}})
-                        : reverse sort keys %{${$diff}->{D}})
-                    : keys %{${$diff}->{D}};
+            map {
+                unshift @stack, [@{$path}, {keys => [$_]}], \${$diff}->{D}->{$_}
+            } $opts{sort}
+                ? ref $opts{sort} eq 'CODE'
+                    ? reverse $opts{sort}(keys %{${$diff}->{D}})
+                    : reverse sort keys %{${$diff}->{D}}
+                : keys %{${$diff}->{D}};
         }
     }
 
     return @list;
 }
 
-=head2 dsplit
+=head2 split_diff
 
 Divide diff to pseudo original structures
 
-    $structs = dsplit($diff);
-    # $structs->{a} - contains items originated from $a
-    # $structs->{b} - same for $b
+    $structs = split_diff(diff($a, $b));
+    # $structs->{a}: items originated from $a
+    # $structs->{b}: same for $b
 
 =cut
 
-sub dsplit($);
-sub dsplit($) {
+sub split_diff($);
+sub split_diff($) {
     my $d = shift;
     _validate_meta($d);
     my $s = {};
@@ -277,13 +295,13 @@ sub dsplit($) {
     if (exists $d->{'D'}) {
         if (ref $d->{'D'} eq 'ARRAY') {
             for my $di (@{$d->{'D'}}) {
-                my $ts = dsplit($di);
+                my $ts = split_diff($di);
                 push @{$s->{'a'}}, $ts->{'a'} if (exists $ts->{'a'});
                 push @{$s->{'b'}}, $ts->{'b'} if (exists $ts->{'b'});
             }
         } else { # HASH
             for my $key (keys %{$d->{'D'}}) {
-                my $ts = dsplit($d->{'D'}->{$key});
+                my $ts = split_diff($d->{'D'}->{$key});
                 $s->{'a'}->{$key} = $ts->{'a'} if (exists $ts->{'a'});
                 $s->{'b'}->{$key} = $ts->{'b'} if (exists $ts->{'b'});
             }
